#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <dos.h>

// screen rows and cols
unsigned __based(__segname("_CODE")) SCR_COLS = 80;
unsigned __based(__segname("_CODE")) SCR_ROWS = 25;

unsigned __based(__segname("_CODE")) CUT_LINES = 0;

char __far * __based(__segname("_CODE")) video_mem = (void __far*)0xB8000000L; 
#define video_page_offset (*(unsigned __far*)MK_FP(0x40, 0x4E))

void outbyte(unsigned port, unsigned char byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [ax];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];  

// Serial port settings

// 2304 == 50 baud
// 1040 == 110 baud
// 768 == 150 baud
// 384 == 300 baud
// 192 == 600 baud
// 96 == 1200 baud
// 48 == 2400 baud
// 24 == 4800 baud
// 12 == 9600 baud
// 6 == 19200 baud
// 3 == 38400 baud
// 2 == 57600 baud
// 1 == 115200 baud
unsigned __based(__segname("_CODE")) BAUD_DIVISOR;
// Parity: 0 = no parity, 1 = odd parity, 3 = even parity
unsigned char PARITY;
// Stop bits
unsigned char STOP;

// What COM port to use. 0 == COM1, 1 == COM2 ...
unsigned char __based(__segname("_CODE")) COM_PORT;

// 2: use RTS/CTS flow control
// 1: use software flow control (XON/XOFF)
// 0: no flow control
unsigned char __based(__segname("_CODE")) FLOW_CONTROL;

#define serialbase ((unsigned short __far*)0x00400000L)
void set_serialport(void)
{
  unsigned mode = 0x3;
  unsigned char old_int_mask;
  mode |= STOP << 2;
  mode |= PARITY << 4;
  
  // enable DLAB
  outbyte(serialbase[COM_PORT]+3, 0x80);
  // set baud rate
  outbyte(serialbase[COM_PORT]+0, BAUD_DIVISOR);
  outbyte(serialbase[COM_PORT]+1, BAUD_DIVISOR>>8);
  // set data bits, parity and stop bits
  outbyte(serialbase[COM_PORT]+3, mode);
  // disable FIFO
  outbyte(serialbase[COM_PORT]+2, 0x00);
  // set interrupt when receiving  
  outbyte(serialbase[COM_PORT]+1, 0x01);
  // set RTS/DSR, IRQ enable
  outbyte(serialbase[COM_PORT]+4, 0x0B);

  old_int_mask = inbyte(0x21);
  if(COM_PORT & 0x1)
  {
    outbyte(0x21, old_int_mask & ~0x08);
  }
  else
  {
    outbyte(0x21, old_int_mask & ~0x10);
  }
}

unsigned char __far * __based(__segname("_CODE")) vram_copy;

volatile unsigned __based(__segname("_CODE")) xoff_received = 0;

int serial_send(unsigned char chr)
{
  if(((FLOW_CONTROL & 1) && xoff_received)
     || ((FLOW_CONTROL & 2) && !(inbyte(serialbase[COM_PORT]+6) & 0x10)))
    return -1;
  // wait until bit 5 is set
  while(!(inbyte(serialbase[COM_PORT]+5) & 0x20));

  outbyte(serialbase[COM_PORT], chr);
  if(inbyte(serialbase[COM_PORT]+5) & 0x02)
    return 1; // overrun
  
  return 0;
}

char __far *i_to_a(unsigned num, char __far *itoa_retstr)
{
  char __far *ptr = itoa_retstr+9;

  *ptr-- = 0;

  do
    *ptr = (num % 10) + '0';
  while((num /= 10) && ptr--);

  return ptr;
}

int send_escape(void)
{
  if(serial_send(0x1B)) return 1;
  if(serial_send('[')) return 1;
  return 0;
}

int send_decimal(unsigned num)
{
  char __far *ptr;
  char itoa_retstr[10];
  for(ptr=i_to_a(num, itoa_retstr);*ptr;ptr++)
    if(serial_send(*ptr)) return 1;
  return 0;
}

//unsigned __based(__segname("_CODE")) prev_diff = -2;
unsigned char __based(__segname("_CODE")) term_cur_row = -1;
unsigned char __based(__segname("_CODE")) term_cur_col = -1;

int send_chr(unsigned char chr)
{
  if(++term_cur_col > SCR_COLS)
  {
    term_cur_col = 1;
    if(term_cur_row != SCR_ROWS - CUT_LINES)
      term_cur_row++;
  }
  return serial_send(chr);
}

int set_cursor_pos(unsigned row, unsigned col)
{
  if(term_cur_row == row && term_cur_col == col) return 0;
//  char itoa_retstr[10];
  if(send_escape()) return 1;
  if(send_decimal(row)) return 1;
  if(serial_send(';')) return 1;
  if(send_decimal(col)) return 1;
  if(serial_send('H')) return 1;

  term_cur_row = row;
  term_cur_col = col;
  return 0;
}

extern void __far * __far old_vector;
extern void __far * __far bios_int_10h;

// color conversion table
unsigned char __based(__segname("_CODE")) ansi_color[] =
{
  0x00,
  0x04,
  0x02,
  0x06,
  0x01,
  0x05,
  0x03,
  0x07
};

unsigned char __based(__segname("_CODE")) color = 0x07;

int set_color(unsigned char color_byte)
{
  if(color_byte & 0x08)
  {
    if(!(color & 0x08)) // intensity on
    {
      if(send_escape()) return 1;
      if(send_decimal(1)) return 1;
      if(serial_send('m')) return 1;
    }
  }
  else if(color & 0x08) // intensity off
  {
    if(send_escape()) return 1;
    if(send_decimal(22)) return 1;
    if(serial_send('m')) return 1;
  }

  // foreground color
  if(send_escape()) return 1;
  if(send_decimal(30 + ansi_color[color_byte&0x7])) return 1;
  if(serial_send('m')) return 1;

  // background color
  if(send_escape()) return 1;
  color_byte >>= 4;
  if(send_decimal(40 + ansi_color[color_byte&0x7])) return 1;
  if(serial_send('m')) return 1;
  return 0;
}

inline int cell_changed(unsigned offset)
{
  const char __far * const ptr1 = video_mem+video_page_offset;
  const char __far * const ptr2 = vram_copy;
  if(ptr1[offset<<1] != ptr2[offset<<1]
     || ptr1[(offset<<1)+1] != ptr2[(offset<<1)+1])
    return 1;
  return 0;
}

unsigned bios_cursor_pos(void);
#pragma aux bios_cursor_pos = \
  "push ax" \
  "push bx" \
  "push cx" \
  "xor bx, bx" \
  "mov ah, 0x3" \
  "int 0x10" \
  "pop cx" \
  "pop bx" \
  "pop ax" \
  value [dx];

/*inline int memcmp(char __far *ptr1, char __far *ptr2, unsigned count)
{
  int rc = 0;
  while(count--)
    if(rc = *ptr1++ - *ptr2++)
      break;
  return rc;
}*/

/*inline int wordcpy(unsigned __far *ptr1, unsigned __far *ptr2, unsigned count)
{
  while(count--)
    *ptr1++ = *ptr2++;
  return 0;
}

inline int wordset(unsigned __far *ptr, unsigned word, unsigned count)
{
  while(count--)
    *ptr++ = word;
  return 0;
}*/
extern void wordcpy(void __far*, void __far*, unsigned);
#pragma aux wordcpy parm [es di] [dx si] [cx];

extern void wordset(void __far*, unsigned, unsigned);
#pragma aux wordset parm [es di] [ax] [cx];

extern int wordcmp(void __far*, void __far *, unsigned);
#pragma aux wordcmp parm [es di] [dx si] [cx] value [cx];

inline int screen_scrolled(void)
{
  unsigned char __far *vram = video_mem+video_page_offset;
  unsigned row;
  unsigned bytes_in_row = SCR_COLS<<1;

  for(row=1;row<SCR_ROWS-1;row++)
  {
    if(!wordcmp(&vram_copy[row*bytes_in_row], vram, (SCR_ROWS-row)*SCR_COLS))
    {
      if(send_escape()) break;
      if(send_decimal(row)) break;
      if(serial_send('S')) break;

      wordcpy((unsigned __far*)vram_copy, (unsigned __far*)vram, (SCR_ROWS-row)*SCR_COLS);

      wordset((unsigned __far*)&vram_copy[(SCR_ROWS-row)*bytes_in_row], -1, row*SCR_COLS);
      return row;
    }
  }

  return 0;
}

inline int first_cell_in_row(unsigned cell)
{
  return !(cell % SCR_COLS);
}

inline unsigned spaces_after_cell(unsigned cell)
{
  unsigned n;
  unsigned __far * ptr = (void __far*)(video_mem+video_page_offset);
  ptr += cell;

  if((char)*ptr != ' ') return 0;

  for(n=1;ptr[n]==*ptr && (cell+n)%SCR_COLS;n++);

  return n;
}

inline int start_line_empty(unsigned __far *cell)
{
  unsigned n;
  unsigned char __far * const ptr = video_mem+video_page_offset;
  if(!first_cell_in_row(*cell))
    return 0;
  if((n = spaces_after_cell(*cell)) < 4)
    return 0;
  if(!wordcmp(&vram_copy[*cell<<1], &ptr[*cell<<1], n))
    return 0;
  set_cursor_pos(term_cur_row, n);
  if(send_escape()) return 1;
  if(send_decimal(1)) return 1;
  if(serial_send('K')) return 1;
  wordcpy((unsigned __far*)&vram_copy[*cell<<1], (unsigned __far*)&ptr[*cell<<1], n);
  *cell += n - 1;

  return 1;
}

inline int end_line_empty(unsigned __far *cell)
{
  unsigned __far * const ptr = (unsigned __far*)(video_mem+video_page_offset);
  unsigned n;
  unsigned line_rem = SCR_COLS - (*cell % SCR_COLS);

  if(!wordcmp((unsigned char __far*)&ptr[(*cell+1)], &vram_copy[(*cell+1)<<1], line_rem))
    return 0;

  if((char)ptr[*cell] != ' ') return 0;
  if(line_rem < 4) return 0;

  for(n=1;(*cell+n) % SCR_COLS;n++)
  {
    if(ptr[*cell+n] != ptr[*cell]) return 0;
  }
  wordcpy((unsigned __far*)&vram_copy[*cell<<1], &ptr[*cell], n);
  *cell += n-1;
  if(send_escape()) return 1;
  if(send_decimal(0)) return 1;
  if(serial_send('K')) return 1;
  set_cursor_pos(term_cur_row, SCR_COLS);

  return 1;
}

void send_screen(void)
{
  unsigned char __far *ptr = video_mem+video_page_offset;
  unsigned n;
  unsigned s;
  int row, col;
  unsigned t = 300/BAUD_DIVISOR;
  unsigned display_size = SCR_COLS*SCR_ROWS;

  if(!(n = wordcmp(ptr, vram_copy, display_size)))
    goto setpos;

  if(!t) t++;

  if(s = screen_scrolled())
    n = s*SCR_COLS;
  else
    n = display_size-n;

  if(n < CUT_LINES*SCR_COLS)
    n = CUT_LINES*SCR_COLS;

//  for(n=CUT_LINES*SCR_COLS;n<SCR_COLS*SCR_ROWS;n++)
  for(;n<display_size;n++)
    if(cell_changed(n))
      goto draw;
  setpos:
  n = bios_cursor_pos();
  col = (n & 0xFF) + 1;
  row = (n >> 8) + 1 - CUT_LINES;
  if(row > 0) set_cursor_pos(row, col);
  return;

  draw:while(t--)
  {
    if(cell_changed(n))
    {
      row = (n / SCR_COLS) + 1 - CUT_LINES;
      col = (n % SCR_COLS) + 1;
      if(set_cursor_pos(row, col))
        return;
      if(ptr[(n<<1)+1] != color)
      {
        if(set_color(ptr[(n<<1)+1]))
          return;
        color = ptr[(n<<1)+1];
      }

      if(!end_line_empty(&n) && !start_line_empty(&n))
      {
        if(send_chr(ptr[n<<1]))
          return;
        vram_copy[n<<1] = ptr[n<<1];
        vram_copy[(n<<1)+1] = ptr[(n<<1)+1];
      }
    }
    if(++n == SCR_COLS*SCR_ROWS) return;//n = CUT_LINES*SCR_COLS;//n = 0;
  }
}

char __far *timer_stack;
extern unsigned __based(__segname("_CODE")) timer_ss;
extern unsigned __based(__segname("_CODE")) timer_sp;

int bios_ungetch(unsigned char c);
#pragma aux bios_ungetch = \
  "mov ah, 0x05" \
  "int 0x16" \
  parm [cx] value [ax];

inline void output2kbdbuff(unsigned char c)
{
  bios_ungetch(c);
}

void __far serial_int(void);
void __far int10h_handler(void);

void serial_int_handler(void)
{
  int c = inbyte(serialbase[COM_PORT]);  

  if(FLOW_CONTROL & 1)
  {
    if(c == 0x13) // XOFF detected
      xoff_received = 1;
    else if(c == 0x11) // XON
      xoff_received = 0;
    else goto savekey;
    return;
  }
  savekey:
  output2kbdbuff(c);
}

void __far timer_int();

void outofmemory(void)
{
  puts("Out of memory");
  exit(-1);
}

int main(char argc, char **argv)
{
  unsigned char ser_int;

  if(argc < 6)
  {
    puts("Usage: $ BIOSCOM [port] [baud divisor] [flow control] [parity]\
 [stop bits] ([rows] [cols] [cut lines] [video memory segment])");
    puts("Port: 0 == COM1, 1 == COM2 ...");
    puts("Flow control: 0 == none, 1 == XON/XOFF, 2 == RTS/CTS, 3 == both");
    puts("Parity: 0 == none, 1 == odd parity, 3 == even parity");
    puts("Stop bits: 0, 1, 2");
    puts("Baud divisor:");
    puts("Value | Baud");
    puts("1       115200");
    puts("2       57600");
    puts("3       38400");
    puts("6       19200");
    puts("12      9600");
    puts("24      4800");
    puts("48      2400");
    puts("You can also use other baud divisors that are supported by your\
 hardware.");
    puts("Rows and cols: How many lines your BIOS terminal has.\
 The defaults are 25 and 80.");
    puts("Cut lines: How many lines to cut from the top. The default is 0.");
    puts("Video memory segment: Set this to B000 for MDA!");
    return 1;
  }

  COM_PORT = atoi(argv[1]);
  BAUD_DIVISOR = atoi(argv[2]);
  FLOW_CONTROL = atoi(argv[3]);
  PARITY = atoi(argv[4]);
  STOP = atoi(argv[5]);

  if(argc >= 8)
  {
    SCR_ROWS = atoi(argv[6]);
    SCR_COLS = atoi(argv[7]);
    printf("Using %u lines and %u columns.\n", SCR_ROWS, SCR_COLS);
  }

  if(argc >= 9)
    CUT_LINES = atoi(argv[8]);

  if(argc >= 10)
    video_mem = MK_FP(strtoul(argv[9], 0, 0x10), 0);

  old_vector = _dos_getvect(0x1C);
  if(!(vram_copy = _fcalloc(4000, 1))) outofmemory();
  if(!(timer_stack = _fmalloc(512))) outofmemory();
  timer_ss = ((unsigned long)timer_stack)>>16;
  timer_sp = ((unsigned)timer_stack)+512;

  set_serialport();

  if(COM_PORT & 0x1) ser_int = 0xB;
  else ser_int = 0xC;

  _dos_setvect(0x1C, timer_int);
  _dos_setvect(ser_int, serial_int);

  puts("Going resident.");

  _dos_keep(0, 0x10+((FP_OFF(main)+0xF)>>4));
  return -1;
}
