#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <dos.h>

void outbyte(unsigned port, unsigned char byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [ax];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];  

#define BAUDS 4800
#define PORT 1

// Serial port settings

// 2304 == 50 baud
// 1040 == 110 baud
// 768 == 150 baud
// 384 == 300 baud
// 192 == 600 baud
// 96 == 1200 baud
// 48 == 2400 baud
// 24 == 4800 baud
// 12 == 9600 baud
// 6 == 19200 baud
// 3 == 38400 baud
// 2 == 57600 baud
// 1 == 115200 baud
#define BAUD_DIVISOR 24
// Parity: 0 = no parity, 1 = odd parity, 3 = even parity
#define PARITY 1
// Stop bits
#define STOP 1

// What COM port to use. 0 == COM1, 1 == COM2 ...
#define COM_PORT 0

// use RTS/CTS flow control
#define FLOW_CONTROL

const unsigned short __far * __far serialbase = (unsigned short __far*)0x00400000L;
void set_serialport(void)
{
  unsigned mode = 0x3;
  mode |= STOP << 2;
  mode |= PARITY << 4;
  
  // enable DLAB
  outbyte(serialbase[COM_PORT]+3, 0x80);
  // set baud rate
  outbyte(serialbase[COM_PORT]+0, BAUD_DIVISOR);
  outbyte(serialbase[COM_PORT]+1, BAUD_DIVISOR>>8);
  // set data bits, parity and stop bits
  outbyte(serialbase[COM_PORT]+3, mode);
  // disable FIFO
  outbyte(serialbase[COM_PORT]+2, 0x00);
  // set interrupt when receiving  
  outbyte(serialbase[COM_PORT]+1, 0x01);
  // set RTS/DSR, IRQ enable
  outbyte(serialbase[COM_PORT]+4, 0x0B);
}

unsigned char __far * __far vram_copy;

volatile unsigned __far xoff_received = 0;

int serial_send(unsigned char chr)
{
  while(xoff_received || !(inbyte(serialbase[COM_PORT]+6) & 0x10));
  // wait until bit 5 is set
  while(!(inbyte(serialbase[COM_PORT]+5) & 0x20));

  outbyte(serialbase[COM_PORT], chr);
  if(inbyte(serialbase[COM_PORT]+5) & 0x02)
    return 1; // overrun
  
  return 0;
}

char __far *i_to_a(unsigned num, char __far *itoa_retstr)
{
  char __far *ptr = itoa_retstr+9;

  *ptr-- = 0;

  do
    *ptr = (num % 10) + '0';
  while((num /= 10) && ptr--);

  return ptr;
}

void set_cursor_pos(unsigned row, unsigned col)
{
  char __far  *ptr;
  char itoa_retstr[10];
  serial_send(0x1B);
  serial_send('[');
  for(ptr=i_to_a(row, itoa_retstr);*ptr;ptr++)
    serial_send(*ptr);
  serial_send(';');
  for(ptr=i_to_a(col, itoa_retstr);*ptr;ptr++)
    serial_send(*ptr);
  serial_send('H');
}

extern void __far * __far old_vector;

void send_screen(void)
{
  unsigned char __far *ptr = (char __far*)0xB8000000;
  static unsigned n = 0;
  unsigned row, col;
  static unsigned prev_diff = -2;
  unsigned t;

  for(n=0;vram_copy[n<<1] == ptr[n<<1];n++)
    if(n == 2000) return;

  for(t=10;t--;)
  {
    if(vram_copy[n<<1] != ptr[n<<1])
    {
      if(prev_diff != n-1)
      {
        row = (n / 80) + 1;
        col = (n % 80) + 1;
        set_cursor_pos(row, col);
      }
      serial_send(ptr[n<<1]);
      vram_copy[n<<1] = ptr[n<<1];
      prev_diff = n;
    }
    if(++n == 2000) n = 0;
  }
}

char __far *timer_stack;
extern unsigned __far timer_ss;
extern unsigned __far timer_sp;

/*void switch2timerstack(void);
#pragma aux switch2timerstack = \
  "mov old_ss, ss" \
  "mov old_sp, ss" \
  "mov ss, timer_ss" \
  "mov sp, timer_sp";

void switch2oldstack(void);
#pragma aux switch2oldstack = \
  "mov ss, old_ss" \
  "mov sp, old_sp";*/

/*void __interrupt timer_int(void)
{
  switch2timerstack();
  send_screen();
  switch2oldstack();
  _chain_intr(old_vector);
}*/

int bios_ungetch(unsigned char c);
#pragma aux bios_ungetch = \
  "mov ah, 0x05" \
  "int 0x16" \
  parm [cx] value [ax];

inline void output2kbdbuff(unsigned char c)
{
/*  unsigned __far *kbd_buffer_head = MK_FP(0x40, 0x1A);
  unsigned __far *kbd_buffer_tail = MK_FP(0x40, 0x1C);
  unsigned char __far *kbd_buffer_pos;

  if(*kbd_buffer_tail - *kbd_buffer_head < 32)
  {
    kbd_buffer_pos = MK_FP(0x40, *kbd_buffer_head);
    *kbd_buffer_pos++ = c;
    *kbd_buffer_pos = 0;
    *kbd_buffer_tail += 2;
    if(*kbd_buffer_tail == 0x1E+32)
      *kbd_buffer_tail -= 32;
  }*/
  bios_ungetch(c);
}

void __interrupt serial_int(void)
{
  int c = inbyte(serialbase[COM_PORT]);  

  if(c == 0x13) // XOFF detected
    xoff_received = 1;
  else if(c == 0x11) // XON
    xoff_received = 0;
  else output2kbdbuff(c);

  outbyte(0x20, 0x20);
}

void __far timer_int();

int main()
{
  old_vector = _dos_getvect(0x1C);
  vram_copy = _fcalloc(4000, 1);
  timer_stack = _fmalloc(512);
  timer_ss = ((unsigned long)timer_stack)>>16;
  timer_sp = ((unsigned)timer_stack)+512;

  set_serialport();

  _dos_setvect(0x1C, timer_int);
  _dos_setvect(0x0C, serial_int);

  _dos_keep(0x10+((FP_OFF(main)+0xF)>>4), 0);

  return 0;
}
